// @todo, split class into 2 - 1 that runs in a window that has a youtube player, and another that doesnt.
// also, maybe we can only inject the iframe code on google / youtube urls / domains to help save cpu.


(async function() {
    'use strict';

    const extensionBaseUrl = document.currentScript.getAttribute('extensionBaseUrl');
    const {leadingAndTrailingFiringThrottle, debounce, watchEx, cmpToPrecision, fromHtml, query, queryAll, MessageType} = await import(extensionBaseUrl + '/yts-util.mjs');
    const {TimePositionMemory} = await import(extensionBaseUrl + '/yts-TimePositionMemory.mjs');

    function YoutubeSeek() {
        // The main dom element that houses the youtube player html.
        // It's also the domain object - the YouTube proprietary player api, with methods.
        // It's not the video element, although the video is a descendant.
        let player;

        // true after .init() was called.
        let initialized = false;

        // Enable/disable application level logging.
        const showLog = false;
        const showWindowId = false;
        const showWindowUrl = false;

        const collectUsageStats = true;

        /** @type {number} */
        let mostRecentSeek;

        const extensionInControlNameMarker = "YoutubeSeek";

        // Note that 1100 is longer than the css animation duration of 1000ms - important so that we don't interrupt the anim.
        const seekButtonAnimationDurationMillis = 1100;

        // Reference to the window object that contains the most recently active youtube player.
        // This is where we send a video seek command to when the user uses keyboard controls.
        // This variable will NOT have a value if this code is executing in the same iframe as the youtube player.
        /** @type {Window} */
        let mostRecentlyActiveWindow;

        let youtubeApiOnStateChangeListener;

        const keyCodes = {
            leftArrow: 37,
            upArrow: 38,
            rightArrow: 39,
            downArrow: 40,
            spaceBar: 32,
        };

        let seekConfig = {
            icons: {}
        };

        let isMac = false;

        // Their api defines constants like YT.PlayerState.ENDED, but I don't know how to access them via my scope.
        // So, I define my own copy of them lol.
        const YoutubePlayerState = {
            "UNSTARTED": -1,
            "ENDED": 0,
            "PLAYING": 1,
            "PAUSED": 2,
            "BUFFERING": 3,
            "VIDEO_CUED": 5
        };

        let playerInitialized = false;

        let throttledStartLookingForPlayerDomElement = leadingAndTrailingFiringThrottle(watchEx(nonThrottledStartLookingForPlayerDomElement), 500);
        let updatePlayer = leadingAndTrailingFiringThrottle(watchEx(nonThrottledUpdatePlayer), 100);

        function startLookingForPlayerDomElement(fromMutationObs) {
            throttledStartLookingForPlayerDomElement(fromMutationObs);
        }

        function isVisible(e) {
            return !!(e.offsetWidth || e.offsetHeight || e.getClientRects().length);
        }

        function isValidVideo(video) {
            // Sometimes there's a video element that has a blank src, so we skip those.
            // They now use multiple video elems per page, even w/ duplicate #player ids!
            // We need to find the active one. They seem to hide the inactive video elements, so
            // we use that as part of our validity check.
            // But, music.youtube.com seems to hide the video element sometimes, even though they're using it to play audio.
            // We should probably do some other checks, like see if its actually playing, but for now we wont.
            return !!(video && video.src && (isVisible(video) || isYoutubeMusicPlayer(video)));
        }

        function findPlayer() {
            const players = queryAll('.html5-video-player, #movie_player').filter(player => {
                const video = player.querySelector('video');
                return isValidVideo(video);
            });
            if (players.length > 1) {
                // Multiple players.
                // If this happens, we need to fix it so we can process them both, or discriminate which one to process.
                recordUsageStats('mp');
            }
            return players[0];
        }

        /**
         *
         * @param {string} selector
         * @return {null|Element}
         */
        function queryPlayerChild(selector) {
            return player ? player.querySelector(selector) : null;
        }

        let numChecks = 0;
        let mutationObserverEnabled = false;
        function nonThrottledStartLookingForPlayerDomElement(fromMutationObs) {
            log(`startLookingForPlayerDomElement fromMutationObs=${fromMutationObs} playerInitialized=${playerInitialized} document.hidden=${document.hidden}`);
            // Prevent potential double init that might happen via queued up setTimeout()s.
            if (playerInitialized) {
                return;
            }

            player = findPlayer();
            if (player) {
                log(`+pf player found. numChecks = ${numChecks}, mutationObserverEnabled = ${mutationObserverEnabled}`, player);
                // @todo remove after confident this doesnt happen
                if (numChecks > 0) {
                    log(`found player on attempt numChecks = ${numChecks}, fromMutationObs = ${fromMutationObs}`);
                }
                playerInitialized = true;
                log(player);
                //attachPlayerStateListener();
                attachPlayerStateListener(player);
                ensureChapterTitleMutationObserverAttached();

                // Disable position indicator for now - not ready yet.
                if (seekConfig.enableTimePositionMemoryForPageReloads || seekConfig.enableTimePositionMemoryForClickedSeekBar) {
                    attachPlayerTimePositionListener(player);
                }

                // Note we only enable this listener after were sure this window has a player.
                ensureWindowActivityListenerAttached();
                ensureKeyListenersAttached();

                attachResizeListener();
                scheduleRenderControls();
                // We don't want to throttle player activity that comes from finding a new player. Ideally, we could
                // fire this like we do now, but also clear the queue of the throttled player so it doesnt fire later, confusing
                // the receiver as to the proper active window.
                sendPlayerActivity();
            }

            numChecks++;
        }

        /**
         * This func helps schedule calls to renderControls(), so that it doesn't interrupt an animation.
         */
        let renderControlsTimerId;
        function scheduleRenderControls() {
            const timeSinceSeek = now() - mostRecentSeek;
            const recentlySeeked = timeSinceSeek < seekButtonAnimationDurationMillis;
            //console.log("time since seek", timeSinceSeek);
            if (recentlySeeked) {
                renderControlsTimerId && clearTimeout(renderControlsTimerId);
                renderControlsTimerId = setTimeout(watchEx(renderControls), seekButtonAnimationDurationMillis);
            } else {
                renderControls();
            }
        }

        /**
         * If the webpage destroys the html element, eg via document.write("<html>new doc</html>")
         * then any listeners attached to that previous document will cease to fire. This can happen with SPAs.
         * edit - can this really happen with document?
         *
         * So, this looks at the element and makes sure our marker property is still present - if not, it attaches a new listener.
         *
         * But, that aside, for efficiency purposes we only want to poll while the document has the user's focus,
         * so we listen to the event and turn our polling interval on or off as needed.
         *
         */
        function ensureVisibilityChangeListenerIsActive() {
            const markerKey = 'YoutubeSeekVisibilityChangeMarker';
            if (!document[markerKey]) {
                document.addEventListener('visibilitychange', watchEx(updatePlayerUpdatingPollingTaskState));
                document[markerKey] = true;
            }
        }

        function updatePlayerUpdatingPollingTaskState() {
            if (!document.hidden) {
                updatePlayerUpdatingPollingTaskState.pollingIntervalTimerId = setInterval(watchEx(() => {
                    updatePlayer();
                }), 1000);
            } else {
                if (updatePlayerUpdatingPollingTaskState.pollingIntervalTimerId) {
                    clearInterval(updatePlayerUpdatingPollingTaskState.pollingIntervalTimerId);
                    updatePlayerUpdatingPollingTaskState.pollingIntervalTimerId = undefined;
                }
            }
        }

        function pollForPlayerUpdates() {
            ensureVisibilityChangeListenerIsActive();
            updatePlayerUpdatingPollingTaskState();
        }

        function reinstallButtonsIfNeeded(useTinySize = false) {
            // First, check if any changes to the buttons or their size is needed (add, remove, or change size).
            // If so, we will just totally uninstall + install them to make the code easier.
            const buttonInfo = getInstalledButtonInfo();
            const needsUninstall = (seekConfig.showShortRewindButton !== buttonInfo.shortRewindButtonInstalled)
                || (seekConfig.showLongRewindButton !== buttonInfo.longRewindButtonInstalled)
                || (seekConfig.showShortForwardButton !== buttonInfo.shortForwardButtonInstalled)
                || (seekConfig.showLongForwardButton !== buttonInfo.longForwardButtonInstalled)
                || (seekConfig.useRectangularStyleButtons !== buttonInfo.useRectangularStyleButtons);

            // todo, we really should just use css classes to toggle show/hide of the buttons, instead of constantly reinstalling them.
            // although, a user who disables all buttons does end up with a cleaner dom this way.
            if (needsUninstall) {
                recordUsageStats('uninst', 1);

                log(`needsUninstall`, buttonInfo, seekConfig);

                uninstallButtons();
                // installSpacerIntoPlayerControls();

                // Now, install what's needed.
                if (seekConfig.showShortRewindButton) {
                    installButtonIntoPlayerControls(useTinySize, true, true);
                }
                if (seekConfig.showLongRewindButton) {
                    installButtonIntoPlayerControls(useTinySize, true, false);
                }
                if (seekConfig.showLongForwardButton) {
                    installButtonIntoPlayerControls(useTinySize, false, false);
                }
                if (seekConfig.showShortForwardButton) {
                    installButtonIntoPlayerControls(useTinySize, false, true);
                }
                setButtonStyle();

                // installSpacerIntoPlayerControls();

            }
        }

        /**
         * This function would be much easier if we could just always
         * uninstall the buttons, then reinstall them in the right config / size.
         * But, we cant because this func often gets called after the user clicks a button
         * (because the player goes into the buffering state, then back to playing state, triggering our event handler), when its spinning.
         * reinstalling it interrupts the animation and feels clunky.
         *
         * So, we try to just update only when needed. But, we need to account for width & config changes.
         */
        function renderControls() {

            log(queryPlayerChild('.ytp-chrome-controls'));
            log(queryPlayerChild('.yts-control-wrapper'));

            if (player) {
                log(`renderControls has player`);
                markExtensionInControl();

                // The youtube music player GUI is very different.
                if (isYoutubeMusicPlayer(player)) {
                    recordUsageStats('ytm', 1);
                    renderControlsForYoutubeMusic();
                    return;
                }

                // Some yt vids hide the controls.
                // We'll skip installing our custom controls, but still let hotkeys work.
                if (!playerHasGuiControls()) {
                    recordUsageStats('noGui', 1);
                    return;
                }

                const videoHasChapters = isChapterTitlesContainerInUse();
                installControlsWrapperIfNotPresent(videoHasChapters);

                if (videoHasChapters) {
                    changeFlexStyleOfChapterTitlesContainer();
                }

                let useTinySize = false;
                // todo maybe measure how much room chapter should have, and decide install.
                // decide button size based on that, not based on title width, because titles change during
                // playback - we need the buttons to stay the same size, as a fraction of player width or space
                // avail w/o chapter element. maybe put chapter flex back to 100% and use fixed width for buttons.
                if (hasRoomToFitShrinkingExpandingControls()) {
                    recordUsageStats('exRoom', 1);
                } else {
                    recordUsageStats('noRoom', 1);
                    log(`cant fit controls, uninstallButtons`, getInstalledButtonInfo());
                    uninstallButtons();
                    // When we can't fit the buttons, it's sometimes because the width is 0 because the player isn't visible. So,
                    // we check and try to find another player on the page.
                    updatePlayer();
                    return;
                }

                reinstallButtonsIfNeeded(useTinySize);
            }
        }

        function renderControlsForYoutubeMusic() {
            let controlsWrapper = getControlWrapper();

            // First, we install it, if its missing.
            if (!controlsWrapper) {
                const middleControlsContainer = query(".ytmusic-player-bar.middle-controls");
                controlsWrapper = fromHtml(`<div class='yts-control-wrapper yts-ytmusic'></div>`);
                middleControlsContainer.insertBefore(controlsWrapper, middleControlsContainer.firstElementChild);
            }

            let useTinySize = false;
            reinstallButtonsIfNeeded(useTinySize);
            return;
        }


        function getButton(direction, duration, controlWrapper) {
            controlWrapper = controlWrapper || getControlWrapper();
            return controlWrapper && controlWrapper.querySelector(`.yts-seek-button-${direction}.yts-seek-button-${duration}`);
        }

        function hasButton(controlWrapper, direction, duration) {
            return !!getButton(direction, duration, controlWrapper);
        }

        // @todo Would be nice if this extracted the seek durations. right now, changing the seek duration in extensions options doesn't
        // cause a button reinstall, although it will seek to the updated value properly.
        function getInstalledButtonInfo() {
            const controlWrapper = getControlWrapper();
            return {
                shortRewindButtonInstalled: hasButton(controlWrapper, 'rewind', 'short'),
                longRewindButtonInstalled: hasButton(controlWrapper, 'rewind', 'medium'),
                shortForwardButtonInstalled: hasButton(controlWrapper, 'forward', 'short'),
                longForwardButtonInstalled: hasButton(controlWrapper, 'forward', 'medium'),
                tinySizeUsed: !!controlWrapper.querySelector('.yts-seek-button.yts-xs'),
                useRectangularStyleButtons: controlWrapper.matches('.yts-use-rectangular-buttons'),
            };
        }

        const wrappedListener = watchEx(onPlayerStateChangeListener);
        function attachPlayerStateListener(player) {
            player.addEventListener("onStateChange", wrappedListener);
        }

        function removePlayerStateListener(player) {
            player.removeEventListener("onStateChange", wrappedListener);
        }

        function attachPlayerTimePositionListener(player) {
            if (!attachPlayerTimePositionListener.attached) {
                attachPlayerTimePositionListener.attached = true;
                let memory = new TimePositionMemory(player, isYoutubeMusicPlayer(player));
                memory.init();
            }
        }

        /**
         * The player sometimes starts in a state where you need to click it before the video will start loading/playing.
         * We can't install our seek buttons/controls until after that click because the native youtube player controls aren't
         * yet shown, and thus their dimensions are not measurable by us yet.
         *
         * So, we wait for the first time the player changes to a state that's associated with the gui controls being shown,
         * and then we render our buttons.
         */
        function onPlayerStateChangeListener(playerStateCode) {
            const statesThatShowControls = [YoutubePlayerState.PAUSED, YoutubePlayerState.PLAYING];
            // const statesThatShowControls = [YoutubePlayerState.BUFFERING, YoutubePlayerState.PAUSED, YoutubePlayerState.PLAYING];
            if (statesThatShowControls.includes(playerStateCode)) {
                scheduleRenderControls();

                // We render again in a moment because it seems youtube can change the width of their controls in this
                // time period, which causes us to run out of room, pushing the floated native youtube controls down where they cant
                // be seen or used.
                setTimeout(watchEx(scheduleRenderControls), 1000);
            }
            log(`YTe event ${playerStateCode}`, arguments, player);
        }

        function ensureWindowActivityListenerAttached() {
            if (!ensureWindowActivityListenerAttached.attached) {
                ensureWindowActivityListenerAttached.attached = true;
                // I don't see the need to send activity too often, but we don't want to throttle for too long, or
                // we might miss legitimate back and forth window focus switching.
                log('ensureWindowActivityListenerAttached');
                const throttledSendPlayerActivity = leadingAndTrailingFiringThrottle(watchEx(sendPlayerActivity), 200);
                document.addEventListener("keydown", watchEx(throttledSendPlayerActivity));
                document.addEventListener("mousedown", watchEx(throttledSendPlayerActivity));
            }
        }

        function attachWindowMessageListener() {
            window.addEventListener("message", watchEx(event => {
                // Skip messages from other peoples code, & messages from our self.
                if (!event || !event.data || !event.data.type || window === event.source || event.data.messageSrcAppName !== 'yts') {
                    return;
                }

                switch (event.data.type) {
                    case MessageType.WINDOW_ACTIVITY:
                        mostRecentlyActiveWindow = event.source;
                        log("got player activity", event);
                        ensureKeyListenersAttached();
                        break;
                    case MessageType.PLAYER_SEEK_COMMAND:
                        log("got seek command", event);
                        dispatchSeek(event.data.seekValue, event.data.buttonInfo);
                        break;
                    case MessageType.TOGGLE_PLAY_PAUSE:
                        log("got toggle play/pause command", event);
                        dispatchPlayPause();
                        break;
                    case MessageType.TOGGLE_MUTE:
                        log("got toggle mute command", event);
                        dispatchMuteUnMute();
                        break;
                    case MessageType.USAGE_STATS:
                        // do nothing.
                        break;
                    default:
                        log("unknown event type",  event);
                }

            }), false);
        }

        // youtube is going in the SPA direction, where navigating is done by just updating the dom instead of loading a new page.
        // This totally screws this extension. For now, we will try to detect this and recover
        // by looking for the updated player.
        function nonThrottledUpdatePlayer() {
            if (playerSeemsValid(player)) {
                return;
            }

            if (player) {
                let newPlayer = findPlayer();
                if (player !== newPlayer) {
                    log(`found new player`, player, newPlayer);
                    recordUsageStats('upd', 1);

                    // Should we even send an activity message here? The parent window already knows we have a player in this window, because
                    // this is a video *update*, not a video first time found. I think if another player is playing, our commands would start
                    // going to this window now, which might confuse the user. But, this video update might happen all by itself, like at the end of a video.
                    // I wouldnt want it to take priority over activity triggered by a deliberate click. Maybe we need to modify the message,
                    // so the receives can make this a low priority activity, and only use it if it has nothing better.
                    // also, hmmm - what if the other vid was just paused by a click, and this vid just finished, which caused it to update to the next vid.
                    // the user might now hear this vid play, and expect their controls go here, instead of the vid they just paused.
                    // arg.... too many possibilites. need to think on it more.
                    sendPlayerActivity();

                    // Remove the event listener from the old player.
                    removePlayerStateListener(player);
                    // Clear the current player so our code will look for one, and find the new one.
                    player = null;
                    playerInitialized = false;
                    startLookingForPlayerDomElement(false);
                }
            } else {
                startLookingForPlayerDomElement(false);
            }
        }

        function playerSeemsValid(player) {
            if (!player) {
                return false;
            }

            const video = player.querySelector('video');
            if (!video || !document.body.contains(video)) {
                return false;
            }

            return isValidVideo(video);
        }

        function sendMessageToTop(obj) {
            log("sendMessageToTop", obj);
            sendPostMessage(window.top, obj);
        }

        function sendSeekToActivePlayer(seekDuration, buttonInfo) {
            if (mostRecentlyActiveWindow) {
                // Maybe we should catch exceptions here?
                // Maybe this window obj no longer exists, eg due to iframe element removal.
                sendPostMessage(mostRecentlyActiveWindow, {type: MessageType.PLAYER_SEEK_COMMAND, seekValue: seekDuration, buttonInfo});
            }
        }

        function sendTogglePlayPauseCommandToActivePlayer() {
            if (mostRecentlyActiveWindow) {
                // Maybe we should catch exceptions here?
                // Maybe this window obj no longer exists, eg due to iframe element removal.
                sendPostMessage(mostRecentlyActiveWindow, {type: MessageType.TOGGLE_PLAY_PAUSE});
            }
        }

        function sendToggleMuteCommandToActivePlayer() {
            if (mostRecentlyActiveWindow) {
                // Maybe we should catch exceptions here?
                // Maybe this window obj no longer exists, eg due to iframe element removal.
                sendPostMessage(mostRecentlyActiveWindow, {type: MessageType.TOGGLE_MUTE});
            }
        }

        function sendPlayerActivity() {
            sendMessageToTop({type: MessageType.WINDOW_ACTIVITY});
        }

        function sendPostMessage(window, msg) {
            const m = Object.assign({}, msg, {messageSrcAppName: 'yts'});
            window.postMessage(m, "*");
        }

        function installButtonIntoPlayerControls(useTinySize, isRewind, isShortSeek) {
            const arrowDirectionWord = isRewind ? 'LeftArrow' : 'RightArrow';
            const arrowDirectionProperty = isRewind ? 'leftArrow' : 'rightArrow';
            const seekDurationWord = isShortSeek ? 'short' : 'medium';
            let hotKeyDescription;
            if (isMac) {
                hotKeyDescription = isShortSeek ? `Ctrl + Command + ${arrowDirectionWord}` : `Ctrl + Shift + ${arrowDirectionWord}`;
            } else {
                hotKeyDescription = isShortSeek ? `Ctrl + ${arrowDirectionWord}` : `Ctrl + Shift + ${arrowDirectionWord}`;
            }
            const hotKeyProperty = `ctrl${isShortSeek ? '' : '.shift'}.${arrowDirectionProperty}`;
            installSeekButton(hotKeyProperty, seekDurationWord, hotKeyDescription, isRewind, useTinySize);
        }

        function uninstallButtons() {
            Array.from(getControlWrapper().childNodes).forEach(el => el.parentNode.removeChild(el));
        }

        function setButtonStyle() {
            const ele = getControlWrapper();
            ele.classList.remove('yts-use-rectangular-buttons', 'yts-use-circular-buttons');
            ele.classList.add(seekConfig.useRectangularStyleButtons ? 'yts-use-rectangular-buttons' : 'yts-use-circular-buttons');
        }

        function getControlWrapper() {
            // On yt music, the controls arent a child of the player element like they are on normal yt.
            return isYoutubeMusicPlayer(player) ? query(".yts-control-wrapper") : queryPlayerChild(".yts-control-wrapper");
        }

        function installControlsWrapperIfNotPresent(videoHasChapters) {
            if (!getControlWrapper()) {
                const timeContainer = queryPlayerChild(".ytp-time-display");
                const wrapper = fromHtml(`<div class='yts-control-wrapper ${videoHasChapters ? 'ytp-chapter-titles' : ''}'></div>`);
                timeContainer.parentNode.insertBefore(wrapper, timeContainer.nextSibling);
            }
        }

        /**
         * They use flex 100% on the container which shows chapter titles, which causes it to take up all remaining space, even though it doesn't need it all to display chapter titles.
         * So, we change the flex style.
         *
         */
        function changeFlexStyleOfChapterTitlesContainer() {
            const chapterContainer = queryPlayerChild(".ytp-chapter-container");
            if (chapterContainer) {

                // On videos w/o chapter titles, we let the controls wrapper shrink via flexbox. But this can be annoying
                // when chapter titles change, causing more/less text to be in the box which changes the size of chapters, and thus changes the
                // size of the buttons. If you're clicking buttons while it changes, you can even suddenly click the wrong button.
                // So, we use fixed sizing for chapter videos, and make it be a fraction of the available space.
                // But, we make sure we give the chapter titles a minimum amount of room so they can at least render a word or
                // two, and after that point we will let our buttons take some of their space away if there isn't much room.
                // It turned out to be easier to set the fixed sizing on the chapter title container. This works because if the chapter wrapper
                // size changes, like when you seek, it puts pressure on the flex items, making them shrink. So we set a max size before hand, so
                // that it already occupies all the space it can get.
                const maxChapterElementWidth = availableWidthInLeftControlAreaIfChapterTitlesWerentThere();
                const maxButtonsWidth = maxWidthNeededToFitShrinkingExpandingControls();
                const chapterContainer = queryPlayerChild(".ytp-chapter-container");

                // We allow a small min width, hopefully displaying a few letters of the chapter, showing the user that it's there, but there's a space problem.
                // This scenario will apply flex pressure, making our buttons small.
                const roomForChapterTitles = Math.ceil(Math.max(80, maxChapterElementWidth - maxButtonsWidth));
                chapterContainer.style.flex = `0 1 ${roomForChapterTitles}px`;
            }
        }

        function isChapterTitlesContainerInUse() {
            const chapterContainer = queryPlayerChild(".ytp-chapter-container");
            if (chapterContainer) {
                const style = window.getComputedStyle(chapterContainer);
                // The chapter container element is always present, but they set display:none for videos w/o any chapters.
                return style && style.display !== 'none';
            }
        }

        function playerHasGuiControls() {
            return queryPlayerChild(".ytp-chrome-controls");
        }

        function isYoutubeMusicPlayer(player) {
            return !!player.closest('ytmusic-player');
        }

        /**
         * Our other extension also controls / seeks video elements. It's generic, meant to work on any website, opposed to this extension,
         * which is dedicated to youtube. So, if both extensions are active, we want this extension to control youtube videos
         * because it's more specialised / better for youtube. So, we mark the dom element, and the other extension looks for this mark
         * so that it wont try to control this specific player element.
         */
        function markExtensionInControl() {
            if (player) {
                const videoElem = getVideoElementFromPlayer();
                if (videoElem) {
                    videoElem.dataset[extensionInControlNameMarker] = "1";
                }
            }
        }

        /**
         * If we have a button installed for this direction and duration of seek command, we send a click event to the button so that it animates.
         * Otherwise, we just seek the player directly.
         *
         * @param {'rewind' | 'forward'} direction
         * @param {'short' | 'medium'} duration
         * @return {boolean}
         */
        function simulateButtonClick(direction, duration) {
            const button = getButton(direction, duration);
            if (!button) {
                return false;
            }

            const event = new MouseEvent('click', {
                view: window,
                bubbles: true,
                cancelable: true
            });

            button.dispatchEvent(event);
            return true;
        }

        function installSeekButton(seekValKey, name, hotKey, isRewind, useTinySize) {
            const seekDuration = getSeekVal(seekValKey);
            const direction = isRewind ? "Rewind" : "Fast Forward";
            const seekDurationDisplay = seekDuration.toFixed(1);
            let title = `${direction} ${seekDurationDisplay} seconds. Hotkey: ${hotKey}. Change this via the Extension options page (click the icon in the top right of your browser).`;
            if (!seekConfig.showSeekButtonTooltips) {
                title = '';
            }
            const buttonHtml = `
            <button class="ytp-button yts-seek-button yts-seek-button-${name} yts-seek-button-${isRewind ? 'rewind' : 'forward'} ${useTinySize ? 'yts-xs' : ''}' aria-label='${title}' title='${title}'">
                <svg version="1.1" width="100%" height="100%" x="0px" y="0px" viewBox="0 0 100 100" xmlns="http://www.w3.org/2000/svg" class="yts-circular-seek-button">
                    <g>
                        <path class="yts-svg-fill ytp-svg-fill" d="M76.961,54.103c0,14.866-12.095,26.961-26.961,26.961S23.039,68.969,23.039,54.103   c0-13.299,9.682-24.372,22.365-26.561v7.806l27.199-15.704L45.404,3.937v8.462C24.422,14.695,8.039,32.519,8.039,54.103   C8.039,77.24,26.863,96.063,50,96.063c23.138,0,41.961-18.823,41.961-41.961H76.961z"></path>
                        <text class="yts-svg-fill ytp-svg-fill yts-svg-duration-text" x="50%" y="55%" dominant-baseline="middle" text-anchor="middle">-X</text>
                    </g>
                </svg>
                <span class="yts-rectangular-seek-button">
                    <span class="yts-rectangular-seek-button-duration">${seekDurationDisplay}</span>
                    <span class="yts-rectangular-seek-button-icon"></span>
                </span>
            </button>
            `;
            const button = fromHtml(buttonHtml);
            getControlWrapper().appendChild(button);
            const path = button.querySelector('path');
            const seekDurationDisplayElem = button.querySelector('text');
            seekDurationDisplayElem.textContent = isRewind ? "-" + seekDuration : "+" + seekDuration;
            const clickSequence = 1;
            const imgRotationInfo = {clickSequence: clickSequence, path, seekDurationDisplayElem, isRewind, rotateDirection: isRewind ? 1 : -1, useTinySize};
            setRotationStyles(imgRotationInfo);

            button.addEventListener("click", watchEx(evt => {
                seekRelative(imgRotationInfo.rotateDirection * -1 * getSeekVal(seekValKey));
                setRotationStyles(imgRotationInfo);

                // On the youtube Music player bar, click events toggle between PiP mode and normal mode, so we stop the events from bubbling
                // to prevent this annoyance when trying to click a seek button.
                evt.stopPropagation();

                // triggerNativeYoutubeSeekIndicator(500, !imgRotationInfo.isRewind);
            }));
        }

        function setRotationStyles(imgRotationInfo) {
            imgRotationInfo.clickSequence++;
            imgRotationInfo.path.style.transform = `scale(${-imgRotationInfo.rotateDirection}, 1) rotate(${360 * imgRotationInfo.clickSequence}deg`;
            imgRotationInfo.seekDurationDisplayElem.style.transform = ` rotate(${360 * imgRotationInfo.clickSequence * -imgRotationInfo.rotateDirection}deg) ${imgRotationInfo.useTinySize ? ' translateY(-1px)' : ''}`;
        }

        let animationTimerId = 0;
        function triggerNativeYoutubeSeekIndicator(animationDuration, isForwardSeek) {
            const ytpBezel = queryPlayerChild('.ytp-bezel');
            const ytpBezelContainer = ytpBezel.parentElement;
            const svg = ytpBezel.querySelector('.ytp-bezel-icon');
            console.log(svg.innerHTML);

            const svgHtmlForward  = `<svg height="100%" version="1.1" viewBox="0 0 36 36" width="100%"><use class="ytp-svg-shadow" xlink:href="#ytp-id-54"></use><path class="ytp-svg-fill" d="m 10,19 c 0,4.4 3.6,8 8,8 4.4,0 8,-3.6 8,-8 h -2 c 0,3.3 -2.7,6 -6,6 -3.3,0 -6,-2.7 -6,-6 0,-3.3 2.7,-6 6,-6 v 4 l 5,-5 -5,-5 v 4 c -4.4,0 -8,3.6 -8,8 z m 6.7,.9 .2,-2.2 h 2.4 v .7 h -1.7 l -0.1,.9 c 0,0 .1,0 .1,-0.1 0,-0.1 .1,0 .1,-0.1 0,-0.1 .1,0 .2,0 h .2 c .2,0 .4,0 .5,.1 .1,.1 .3,.2 .4,.3 .1,.1 .2,.3 .3,.5 .1,.2 .1,.4 .1,.6 0,.2 0,.4 -0.1,.5 -0.1,.1 -0.1,.3 -0.3,.5 -0.2,.2 -0.3,.2 -0.5,.3 C 18.3,22 18.1,22 17.9,22 17.7,22 17.5,22 17.4,21.9 17.3,21.8 17.1,21.8 16.9,21.7 16.7,21.6 16.7,21.5 16.6,21.3 16.5,21.1 16.5,21 16.5,20.8 h .8 c 0,.2 .1,.3 .2,.4 .1,.1 .2,.1 .4,.1 .1,0 .2,0 .3,-0.1 L 18.4,21 c 0,0 .1,-0.2 .1,-0.3 v -0.6 l -0.1,-0.2 -0.2,-0.2 c 0,0 -0.2,-0.1 -0.3,-0.1 h -0.2 c 0,0 -0.1,0 -0.2,.1 -0.1,.1 -0.1,0 -0.1,.1 0,.1 -0.1,.1 -0.1,.1 h -0.6 z" id="ytp-id-54"></path></svg>`;
            const svgHtmlBackward = `<svg height="100%" version="1.1" viewBox="0 0 36 36" width="100%"><use class="ytp-svg-shadow" xlink:href="#ytp-id-65"></use><path class="ytp-svg-fill" d="M 18,11 V 7 l -5,5 5,5 v -4 c 3.3,0 6,2.7 6,6 0,3.3 -2.7,6 -6,6 -3.3,0 -6,-2.7 -6,-6 h -2 c 0,4.4 3.6,8 8,8 4.4,0 8,-3.6 8,-8 0,-4.4 -3.6,-8 -8,-8 z m -1.3,8.9 .2,-2.2 h 2.4 v .7 h -1.7 l -0.1,.9 c 0,0 .1,0 .1,-0.1 0,-0.1 .1,0 .1,-0.1 0,-0.1 .1,0 .2,0 h .2 c .2,0 .4,0 .5,.1 .1,.1 .3,.2 .4,.3 .1,.1 .2,.3 .3,.5 .1,.2 .1,.4 .1,.6 0,.2 0,.4 -0.1,.5 -0.1,.1 -0.1,.3 -0.3,.5 -0.2,.2 -0.3,.2 -0.4,.3 C 18.5,22 18.2,22 18,22 17.8,22 17.6,22 17.5,21.9 17.4,21.8 17.2,21.8 17,21.7 16.8,21.6 16.8,21.5 16.7,21.3 16.6,21.1 16.6,21 16.6,20.8 h .8 c 0,.2 .1,.3 .2,.4 .1,.1 .2,.1 .4,.1 .1,0 .2,0 .3,-0.1 L 18.5,21 c 0,0 .1,-0.2 .1,-0.3 v -0.6 l -0.1,-0.2 -0.2,-0.2 c 0,0 -0.2,-0.1 -0.3,-0.1 h -0.2 c 0,0 -0.1,0 -0.2,.1 -0.1,.1 -0.1,0 -0.1,.1 0,.1 -0.1,.1 -0.1,.1 h -0.7 z" id="ytp-id-65"></path></svg>`;
            const svgButton = `
                <svg version="1.1" width="100%" height="100%" x="0px" y="0px" viewBox="0 0 100 100" xmlns="http://www.w3.org/2000/svg">
                    <g>
                        <path class="yts-svg-fill ytp-svg-fill" d="M76.961,54.103c0,14.866-12.095,26.961-26.961,26.961S23.039,68.969,23.039,54.103   c0-13.299,9.682-24.372,22.365-26.561v7.806l27.199-15.704L45.404,3.937v8.462C24.422,14.695,8.039,32.519,8.039,54.103   C8.039,77.24,26.863,96.063,50,96.063c23.138,0,41.961-18.823,41.961-41.961H76.961z"></path>
                        <text class="yts-svg-fill ytp-svg-fill yts-svg-duration-text" x="50%" y="55%" dominant-baseline="middle" text-anchor="middle">-X</text>
                    </g>
                </svg>
        `;

            resetAnimation(ytpBezel);

            // svg.innerHTML = isForwardSeek ? svgHtmlForward : svgHtmlBackward;
            svg.innerHTML = svgButton;
            ytpBezelContainer.style.display = 'block';

            // If an anim is running, we will stop the timer, because were about to start a new anim,
            // and dont want the old timer to stop the new anim prematurely.
            if (animationTimerId) {
                clearTimeout(animationTimerId);
            }
            animationTimerId = setTimeout(() => {
                ytpBezelContainer.style.display = 'none';
                animationTimerId = 0;
            }, animationDuration);
        }

        function resetAnimation(elem) {
            elem.style.animation = 'none';
            elem.offsetHeight; /* trigger reflow */
            elem.style.animation = null;
        }

        function seekRelative(offset) {
            const elem = getVideoElementFromPlayer();
            elem && seekRelativeNew(elem, offset, elem.loop, true);
            mostRecentSeek = now();
        }

        /**
         * @return {null|HTMLMediaElement}
         */
        function getVideoElementFromPlayer() {
            return queryPlayerChild('video');
        }

        function now() {
            return new Date().getTime();
        }

        /**
         *
         * @param {HTMLMediaElement} media
         * @param {number} seconds
         * @param {boolean} wrap
         * @param {boolean} fastSeek
         */
        function seekRelativeNew(media, seconds, wrap, fastSeek) {
            let newPos;
            let duration = media.duration;
            let currentTime = media.currentTime;
            const isAtEnd = cmpToPrecision(duration, currentTime, 3) === 0;
            const infinitesimal = 0.001;

            if (wrap) {
                newPos = (currentTime + duration + seconds) % duration;
            } else {
                newPos = Math.max(infinitesimal, Math.min(duration - infinitesimal, currentTime + seconds));
            }

            if (fastSeek && media.fastSeek) {
                media.fastSeek(newPos);
                log(`fs`);
            } else {
                media.currentTime = newPos;
            }

            // If the user seeks to the end, the video stops playing. Even if they then seek backwards by a bit,
            // the video stays paused. But, this sucks - if were at the end and seek backwards, I think  it should start playing again.
            if (isAtEnd) {
                media.play();
            }
        }

        /**
         * If this code is executing in the same iframe as the youtube player, then this func will call it's api to seek.
         * Otherwise, it will send a cross window communication event to tell the code in the correct iframe to seek.
         *
         * @param {number} offset
         * @param {Object} buttonInfo
         * @param {'rewind' | 'forward'} buttonInfo.direction
         * @param {'short' | 'medium'} buttonInfo.duration
         */
        function dispatchSeek(offset, buttonInfo) {
            if (player) {
                if (buttonInfo) {
                    simulateButtonClick(buttonInfo.direction, buttonInfo.duration);
                } else {
                    seekRelative(offset);
                }
                recordUsageStats('seek', 1);
            } else if (mostRecentlyActiveWindow) {
                sendSeekToActivePlayer(offset, buttonInfo);
            }
        }

        function dispatchPlayPause() {
            if (player) {
                togglePlayPause();
            } else if (mostRecentlyActiveWindow) {
                sendTogglePlayPauseCommandToActivePlayer();
            }
        }

        function togglePlayPause() {
            if (!player) {
                return;
            }

            // todo, maybe also allow buffering state?
            if (player.getPlayerState() === YoutubePlayerState.PLAYING) {
                player.pauseVideo();
            } else {
                player.playVideo();
            }
        }

        function dispatchMuteUnMute() {
            if (player) {
                toggleMuteUnMute();
            } else if (mostRecentlyActiveWindow) {
                sendToggleMuteCommandToActivePlayer();
            }
        }

        function toggleMuteUnMute() {
            if (!player) {
                return;
            }

            if (player.isMuted()) {
                player.unMute();
            } else {
                player.mute();
            }
        }

        function getMinWidthPerButton() {
            // 10 is an arbitrary number, and represents a sane visual minimum, and will result in a very small button that has been shrunken down.
            return 10;
        }

        function getFullWidthPerButton() {
            return isPlayerFullScreen() ? 54 : 36;
        }

        function sumWidthOfChildren(elem) {
            return Array.from(elem.children).reduce((accum, childElem) => {
                return accum + Math.ceil(childElem.getBoundingClientRect().width);
            }, 0);
        }

        function availableWidthAmongChildren(parentElem) {
            return parentElem.clientWidth - sumWidthOfChildren(parentElem);
        }

        function numButtonsDesiredToInstall() {
            return [
                seekConfig.showShortRewindButton,
                seekConfig.showLongRewindButton,
                seekConfig.showShortForwardButton,
                seekConfig.showLongForwardButton
            ].filter(val => !!val).length;
        }

        // Min width needed to fit a very shrunken set of buttons.
        function minWidthNeededToFitShrinkingExpandingControls() {
            return getMinWidthPerButton() * numButtonsDesiredToInstall();
        }

        // Basically how wide the button assembly will be at its full, non-shrunken size.
        function maxWidthNeededToFitShrinkingExpandingControls() {
            // approx sizes
            const shortButtonWidth =  isPlayerFullScreen() ? 50 : 42;
            const longButtonWidth =  isPlayerFullScreen() ? 50 : 28;
            return (seekConfig.showShortRewindButton * shortButtonWidth)
                + (seekConfig.showLongRewindButton * longButtonWidth)
                + (seekConfig.showShortForwardButton * shortButtonWidth)
                + (seekConfig.showLongForwardButton * longButtonWidth);
        }

        function chapterTitlesHaveRoomToSpareBySqueezingThings() {
            // Some chapter titles are long, and so take up all available room.
            // Our users want the buttons, and so we balance that against their ability to see the entire chapter title.
            // As long as the chapter title is at least X px long, we will add out buttons because they will still
            // see an abbreviated chapter title after out buttons take some of their space away due to the flexbox layout.
            log(`chapterTitlesHaveRoomToSpareBySqueezingThings widthOfChapterTitleContainer=${widthOfChapterTitleContainer()} > ${minWidthWeShouldAllowChapterTitlesToBe()}`);
            return widthOfChapterTitleContainer() > minWidthWeShouldAllowChapterTitlesToBe();
        }

        function widthOfChapterTitleContainer() {
            return queryPlayerChild(".ytp-chapter-container").clientWidth;
        }

        function availableWidthInChromeControlArea() {
            // The 3 children should be .ytp-left-controls and .ytp-right-controls, (and sometimes .ytp-chapter-container )
            return availableWidthAmongChildren(queryPlayerChild(".ytp-chrome-controls")) + existingCustomControlsWidth();
        }

        function availableWidthInLeftControlArea() {
            return availableWidthAmongChildren(queryPlayerChild(".ytp-left-controls")) + existingCustomControlsWidth();
        }

        function availableWidthInLeftControlAreaIfChapterTitlesWerentThere() {
            // The chapter titles container is shrink to fit, so its current size depends on how much text is inside.
            // Were trying to calculate the max size it *could be* if it was full.
            // @todo, they use overflow elipses, and we need to check if the size keeps increasing as it overflows.
            return availableWidthInLeftControlArea() + widthOfChapterTitleContainer();
        }

        function availableWidthInLeftControlAreaIfChapterTitlesWerentThere() {
            log(`left-------------------`);
            const left = queryPlayerChild(".ytp-left-controls");
            log(`width=${left.getBoundingClientRect().width} elem=${left.classList}`);
            const childWidth = Array.from(left.children).reduce((accum, childElem) => {
                log(`width=${childElem.getBoundingClientRect().width} elem=${childElem.classList}`);
                return accum + Math.ceil(childElem.getBoundingClientRect().width);
            }, 0);
            const fin = left.getBoundingClientRect().width - childWidth + widthOfChapterTitleContainer() + existingCustomControlsWidth();
            log(`fin=${fin}`);
            log(`left-------------------/////`);
            return fin;
        }

        function minWidthWeShouldAllowChapterTitlesToBe() {
            return 80;
        }

        function isPlayerFullScreen() {
            return player.matches('.ytp-fullscreen');
        }

        function availableWidthInControlArea() {
            // log(`availableWidthInControlArea Math.max(${availableWidthInLeftControlArea()}, right=${availableWidthInChromeControlArea()}) = ${Math.max(availableWidthInLeftControlArea(), availableWidthInChromeControlArea())}`);
            // Youtube changed their layout back and forth, making us need to consider 2 possible GUI layouts they might use.
            // In one version, the .ytp-left-controls is a float, and tends to be only 200px wide. So, there's open space to the right side of it in a sibling position.
            // The other version, .ytp-left-controls uses display:flex and so will grow wide, taking up all available space in its parent.
            // When this happens, the free space should be inside of it.
            // So, we measure free space both inside of it, and inside its parent, and use the larger one scenario.
            // But, in either case, we always inject our custom controls to be a child of .ytp-left-controls.
            return Math.max(availableWidthInLeftControlArea(), availableWidthInChromeControlArea());
        }

        // TODO check if left has any room, according to its children. this is for the flex case, because the left outside dims will be full width.
        // also check how much room .ytp-chrome-controls has.

        function existingCustomControlsWidth() {
            const wrapper = getControlWrapper();
            return wrapper ? wrapper.clientWidth : 0;
        }

        function hasRoomToFitShrinkingExpandingControls() {
            log(`hasRoomToFitShrinkingExpandingControls needed=${minWidthNeededToFitShrinkingExpandingControls()} < avail=${availableWidthInControlArea()}`);
            return minWidthNeededToFitShrinkingExpandingControls() < availableWidthInControlArea() || chapterTitlesHaveRoomToSpareBySqueezingThings();
        }

        function getSeekVal(key) {
            return parseFloat(seekConfig[key] || 1);
        }

        /**
         * We only want to setup our key listeners on pages where they're needed, which is windows with a youtube player, or windows
         * with a descendant youtube player. Our use of evt.stopImmediatePropagation()  in our keydown handler is fairly
         * intrusive, and so we don't want to do it on pages without a player.
         *
         * We still listen for keys and messages on non-youtube.com pages because this code needs to run in
         * parent pages that embed the youtube player via iframe. This allows us to send hotkeys to the child iframe
         * even if the iframe doesnt currently have focus.
         */
        function ensureKeyListenersAttached() {
            // This var will only be set if we got an activity message, meaning there's a player in this window, or a descendent window.
            if (!currentWindowIsAwareOfAPlayer()) {
                return;
            }
            if (!ensureKeyListenersAttached.attached) {
                log(`attachKeyListener`);
                attachKeyListener();
                ensureKeyListenersAttached.attached = true;
            }
        }

        function attachKeyListener() {
            function recordHotkeyUsage() {
                recordUsageStats('hk', 1);
            }

            // If they didnt configure hotkeys to work, we return to prevent adding the key listener.
            const enableHotkeys = (seekConfig.enableYoutubeComHotkeys && currentDomainIsYoutube()) || (seekConfig.enableNonYoutubeComHotkeys && !currentDomainIsYoutube());
            if (!enableHotkeys) {
                return;
            }

            const handler = watchEx(evt => {

                if (evt.ctrlKey) {
                    if (evt.shiftKey) {
                        if (evt.which === keyCodes.leftArrow) {
                            recordHotkeyUsage();
                            evt.stopImmediatePropagation();
                            dispatchSeek(-getSeekVal("ctrl.shift.leftArrow"), {direction: 'rewind', duration: 'medium'});
                        } else if (evt.which === keyCodes.rightArrow) {
                            recordHotkeyUsage();
                            evt.stopImmediatePropagation();
                            dispatchSeek(getSeekVal("ctrl.shift.rightArrow"), {direction: 'forward', duration: 'medium'});
                        } else if (evt.which === keyCodes.downArrow) {
                            recordHotkeyUsage();
                            evt.stopImmediatePropagation();
                            // Seek to start: this will figure out the max/min position.
                            dispatchSeek(-1000000);
                        } else if (evt.which === keyCodes.upArrow) {
                            recordHotkeyUsage();
                            evt.stopImmediatePropagation();
                            // Seek to end: this will figure out the max/min position.
                            dispatchSeek(1000000);
                        } else if (evt.which === keyCodes.spaceBar) {
                            recordHotkeyUsage();
                            evt.stopImmediatePropagation();
                            dispatchPlayPause();
                        }
                    } else {
                        if (evt.which === keyCodes.leftArrow) {
                            recordHotkeyUsage();
                            evt.stopImmediatePropagation();
                            dispatchSeek(-getSeekVal("ctrl.leftArrow"), {direction: 'rewind', duration: 'short'});
                        } else if (evt.which === keyCodes.rightArrow) {
                            recordHotkeyUsage();
                            evt.stopImmediatePropagation();
                            dispatchSeek(getSeekVal("ctrl.rightArrow"), {direction: 'forward', duration: 'short'});
                        } else if (evt.which === keyCodes.spaceBar) {
                            recordHotkeyUsage();
                            evt.stopImmediatePropagation();

                            if (evt.altKey) {
                                dispatchMuteUnMute();
                            } else {
                                dispatchPlayPause();
                            }
                        }
                    }
                }

            });

            // We need to use the capturing phase because the yt-hotkey-manager also listens for keydown events in the capturing phase,
            // and they seek +- 5s on left/right arrow, conflicting with our seek hotkeys.
            // So, we stop propagation of the event to prevent their handler from seeking too.
            document.addEventListener("keydown", handler, true);
        }

        /**
         * When the window is resized, the youtube player width adjusts accordingly, and we may no longer have enough room to
         * fit our controls in the youtube player, so we need to check. Otherwise, we will push
         * some of the native controls out of view.
         *
         * The resize event can happen rapidly, so we need to throttle it.
         */
        function attachResizeListener() {
            if (!attachResizeListener.attached) {
                attachResizeListener.attached = true;
                const throttledRenderFunc = leadingAndTrailingFiringThrottle(watchEx(renderControls), 600);
                const debouncedRenderFunc = debounce(watchEx(renderControls), 1000);
                window.addEventListener("resize", evt => {
                    log('resize');
                    throttledRenderFunc();

                    // We re-render in a moment because sometimes measurements aren't correct, resulting in the wrong size buttons being displayed.
                    // I think this is because a resize is a very resource intensive time, and so waiting a bit gives things time to settle down, allowing
                    // the browser to get caught up on executing code or rendering / resizing etc...
                    // We debounce this func because we really only want it to fire when the user is finished
                    // resizing - we dont need it to fire while they're still making adjustments (although, it wont hurt, but we just try to keep cpu use low).
                    debouncedRenderFunc();

                }, false);
            }
        }

        function ensureChapterTitleMutationObserverAttached() {
            const chap = queryPlayerChild('.ytp-chapter-container');
            const attrName = 'youtubeSeekMutationObserverEnabled';
            if (!chap || chap.getAttribute(attrName)) {
                return;
            }
            chap.setAttribute(attrName, 'true')

            function callback(mutations) {
                if (mutations.length > 0) {
                    // When the chapter title text changes, this changes how much room we have to add our buttons
                    // so we need to rerender.
                    scheduleRenderControls();
                    log('chapter text changed');
                }
            }

            const cb = watchEx(callback);
            // const cb = leadingAndTrailingFiringThrottle(watchEx(callback), 500);
            let observer = new MutationObserver(cb);
            observer.observe(chap, {subtree: true, childList: true, characterData: true});
        }

        function isTopWindow() {
            return window === window.top;
        }

        function currentDomainIsYoutube() {
            if (currentDomainIsYoutube.cachedValue === undefined) {
                currentDomainIsYoutube.cachedValue = window.location.hostname.endsWith('youtube.com') || window.location.hostname.endsWith('youtube-nocookie.com');
            }

            return currentDomainIsYoutube.cachedValue;
        }

        // Meaning there's a player in this window, or a descendent window
        function currentWindowIsAwareOfAPlayer() {
            return !!player || !!mostRecentlyActiveWindow;
        }

        function log() {
            if (!showLog) {
                return;
            }

            // This id is helpful to differentiate which window a log event is coming from - there's often many different windows involved when
            // running this extension. As such, a separate copy of this code runs in each window.
            window.youtubeSeekWindowId = window.youtubeSeekWindowId || (new Date().getTime() - 1483925683061 - 265597);

            let logIdHelper = isTopWindow() ? "top" : "sub";
            if (showWindowId) {
                logIdHelper += window.youtubeSeekWindowId;
            }
            if (showWindowUrl) {
                logIdHelper += location.href;
            }

            const args = [logIdHelper].concat(Array.from(arguments));
            console.log.apply(console, args);
        }

        // let send = cusTh
        let throttledSendAccumulatedUsageStats = leadingAndTrailingFiringThrottle(watchEx(sendAccumulatedUsageStats), 2000);
        let accumulatedUsageStats = {};
        function sendAccumulatedUsageStats() {
            collectUsageStats && sendMessageToTop({type: MessageType.USAGE_STATS, usageStats: accumulatedUsageStats});

            // Now that we've sent it, clear out the aggregated values.
            Object.keys(accumulatedUsageStats).forEach(key => {
                delete accumulatedUsageStats[key];
            });
        }

        /**
         * Send the key + value to our content script in the parent window, where we can record the usage stats into chromes storage.
         *
         * For performance, we want to reduce how many window.postMessage() and resulting chrome storage calls we make.
         * So, this func will aggregate the usage stats, and send 1 big message every so often.
         *
         * @param key
         * @param incrementalValue
         */
        function recordUsageStats(key, incrementalValue) {
            accumulatedUsageStats[key] = accumulatedUsageStats[key] || 0;
            accumulatedUsageStats[key] += incrementalValue;

            collectUsageStats && throttledSendAccumulatedUsageStats();
        }

        this.init = watchEx(function() {
            // We always attach a message listener, because we always send seek commands through a message, even to our own window.
            // But, we also need it so we get messages from child iframes that have a player.
            attachWindowMessageListener();

            // We need to find a better way to detect player changes, but youtube changes their code so often
            // that it's always breaking us. So, for now, on youtube.com pages we will just poll and look for an updated
            // player every few seconds. The player seems to change when the user navigates within youtube, like going between a
            // video page and a channel page.
            if (currentDomainIsYoutube()) {
                updatePlayer();
                pollForPlayerUpdates();
                recordUsageStats('init', 1);
            }

            log(`init yts`, this);
        });

        this.recordError = function() {
            recordUsageStats('err', 1);
        };

        this.setConfig = watchEx(function(cfg) {
            // log(cfg);
            seekConfig = cfg;
            isMac = seekConfig.os === 'mac';
            if (!initialized) {
                this.init();
                initialized = true;
            }
            // Need to update the controls so they use the newest config values.
            scheduleRenderControls();
        });

        (function construct() {
            // Expose the player as a read only attribute.
            Object.defineProperty(this, 'player', {
                get: () => player,
                enumerable: true
            });
        }).call(this);
    }
    YoutubeSeek.numErrors = 0;

    const ys = new YoutubeSeek();
// We call init only after we receive the config via setConfig().
// ys.init();
    window.YoutubeSeek = ys;

})();

