/**
 * This script injects our js and css into the webpage after it loads. It also
 * serves as a communication point between the webpage and the extension, allowing the
 * webpage to update if the user changes their settings.
 */


const showLog = false;


(async () => {
    const {getSavedOptions} = await import(chrome.extension.getURL('/saved-settings-misc.mjs'));
    const {MessageType} = await import(chrome.extension.getURL('/yts-util.mjs'));
    const {ChromeStorageUpdateSynchronizer} = await import(chrome.extension.getURL('/ChromeStorageUpdateSynchronizer.mjs'));
    const updateSynchronizer = new ChromeStorageUpdateSynchronizer(chrome.storage.local);


    injectFile(chrome.extension.getURL('/yts-controls.js'), "script", {name: 'yts-script', attributes: {extensionBaseUrl: getExtensionBaseUrl()}}).then(sendUpdatedConfigToWebPage);

    // injectFile(chrome.extension.getURL('yts-TimePositionMemory.mjs'), "script", {attributes: {type: "module"}}).then(() => {
    //     setTimeout(sendUpdatedConfigToWebPage, 1100);
    // });

    // When the settings/config changes, we inject it into the webpage so it can update without the user needing to reload the page.
    chrome.storage.onChanged.addListener(function(changes, namespace) {
        // We only need to update the page if certain config changed. We update the usageStats key very often,
        // but the page doesn't need to be updated for that, so we only update if something else changed.
        const keys = Object.keys(changes);
        const onlyUsageStatsChanged = keys.length === 1 && keys.includes('usageStats');
        if (!onlyUsageStatsChanged) {
            log('######################################## usageStats');
            sendUpdatedConfigToWebPage();
        }
    });

    window.addEventListener("message", function(event) {
        // Skip messages from other peoples code.
        if (!event || !event.data || !event.data.type || event.data.messageSrcAppName !== 'yts') {
            return;
        }

        const d = event.data;
        if (d && d.type === MessageType.USAGE_STATS && d.usageStats) {
            log('USAGE_STATS update', d);
            updateLocalStorage(d.usageStats);
        }
    }, false);

    function isTopWindow() {
        return window === window.top;
    }

    function log() {
        const args = [isTopWindow() ? "top" : "sub", location.href].concat(arguments);
        showLog && console.log.apply(console, args);
    }

    function sendUpdatedConfigToWebPage() {
        Promise.all([getSavedOptions(), getSavedLocalStorage()]).then(values => {
            let syncStorage = values[0];
            let localStorage = values[1];
            let config = Object.assign({}, syncStorage, localStorage);

            // We inject a script using a data url to achieve a dynamic script.
            // Need to do this junk because of chrome content script isolated environment.
            injectFile(scriptDataUrl(scriptContentCallback(config)), "script");
        });
    }

    function removeElement(ele) {
        if (ele && ele.parentNode) {
            ele.parentNode.removeChild(ele);
        }
    }

    function removeElementAsync(ele) {
        setTimeout(() => {
            removeElement(ele);
        }, 15);
    }

    // Returns a url like chrome-extension://jiipohchomedbgnhmebkjeejlmhjlbil
    function getExtensionBaseUrl() {
        const file = '/yts-controls.js';
        const url = chrome.extension.getURL(file);
        if (!url.endsWith(file)) {
            throw new Error(`unexpected extension url='${url}' doesnt contain file='${file}'`);
        }
        return url.slice(0, url.length - file.length);
    }

    function injectFile(url, type, options = {}) {
        return new Promise(function (resolve, reject) {
            const elem = document.createElement(type);
            switch (type) {
                case 'link':
                    elem.href = url;
                    elem.rel  = 'stylesheet';
                    elem.type = 'text/css';
                    break;
                case 'script':
                    elem.src = url;
                    break;
                default:
                    throw new Error("Unhandled switch case:" + type);
            }

            if (options.attributes) {
                Object.keys(options.attributes).forEach(key => {
                    elem.setAttribute(key, options.attributes[key]);
                });
            }

            // elem.setAttribute('parentExtensionId', chrome.runtime.id);

            elem.onload = function() {
                resolve();
                if (options.autoRemoveElement) {
                    removeElementAsync(elem);
                }
            };
            elem.onerror = () => {
                reject();
                removeElementAsync(elem);
            };

            (document.head || document.documentElement).appendChild(elem);
        });
    }

    function scriptDataUrl(scriptContents) {
        return "data:text/javascript;base64," + btoa(scriptContents);
    }

    function scriptContentCallback(configArg) {
        // We inject this js code fragment into the webpage potentially before the code which defines and creates an
        // instance of the YoutubeSeek class completes. So, we make this code keep trying until the object gets instantiated.
        return `(function eventuallySetConfig() {
            if (window.YoutubeSeek) window.YoutubeSeek.setConfig(${JSON.stringify(configArg)});
            else setTimeout(eventuallySetConfig, 500);
        })();`;
    }

    function updateLocalStorage(incrementalKeyValues) {
        updateSynchronizer.updateKeys(['usageStats'], localVals => {
            localVals.usageStats = localVals.usageStats || {};

            Object.entries(incrementalKeyValues).forEach(pair => {
                let [key, incrementalValue] = pair;
                localVals.usageStats[key] = localVals.usageStats[key] || 0;
                localVals.usageStats[key] += incrementalValue;
            });

            return new Promise(resolve => {
                chrome.storage.local.set(localVals, resolve);
            });
        });
    }

    function getSavedLocalStorage() {
        return new Promise(resolve => {
            chrome.storage.local.get({os: 'n/a'}, resolve);
        });
    }

})();




